@@ -3,7 +3,7 @@
 Simple Smith-Waterman aligner
 '''
 import sys
-import StringIO
+from io import StringIO
 
 
 class ScoringMatrix(object):
@@ -21,7 +21,7 @@ class ScoringMatrix(object):
     Rows and Columns must be in the same order
 
     '''
-    def __init__(self, filename=None, text=None):
+    def __init__(self, filename=None, text=None, wildcard_score=0):
         assert filename or text
 
         if filename:
@@ -31,9 +31,10 @@ class ScoringMatrix(object):
 
         self.scores = []
         self.bases = None
+        self.wildcard_score = wildcard_score
 
         for line in fs:
-            if line[0] == '#':
+            if line[0] == '#' or not line.strip():
                 continue
 
             if not self.bases:
@@ -41,11 +42,14 @@ class ScoringMatrix(object):
                 self.base_count = len(self.bases)
             else:
                 cols = line.split()
-                self.scores.extend([int(x) for x in cols[1:]])
+                self.scores.extend([float(x) for x in cols[1:]])
 
         fs.close()
 
-    def score(self, one, two):
+    def score(self, one, two, wildcard=None):
+        if self.wildcard_score and wildcard and (one in wildcard or two in wildcard):
+            return self.wildcard_score
+
         one_idx = 0
         two_idx = 0
         for i, b in enumerate(self.bases):
@@ -57,16 +61,21 @@ class ScoringMatrix(object):
         return self.scores[(one_idx * self.base_count) + two_idx]
 
 
-class NucleotideScoringMatrix(object):
+class IdentityScoringMatrix(object):
     def __init__(self, match=1, mismatch=-1):
         self.match = match
         self.mismatch = mismatch
 
-    def score(self, one, two):
+    def score(self, one, two, wildcard=None):
+        if wildcard and (one in wildcard or two in wildcard):
+            return self.match
+
         if one == two:
             return self.match
         return self.mismatch
 
+NucleotideScoringMatrix = IdentityScoringMatrix
+
 
 class Matrix(object):
     def __init__(self, rows, cols, init=None):
@@ -82,7 +91,7 @@ class Matrix(object):
 
 
 class LocalAlignment(object):
-    def __init__(self, scoring_matrix, gap_penalty=-1, gap_extension_penalty=-1, gap_extension_decay=0.0, prefer_gap_runs=True, verbose=False, globalalign=False):
+    def __init__(self, scoring_matrix, gap_penalty=-1, gap_extension_penalty=-1, gap_extension_decay=0.0, prefer_gap_runs=True, verbose=False, globalalign=False, wildcard=None, full_query=False):
         self.scoring_matrix = scoring_matrix
         self.gap_penalty = gap_penalty
         self.gap_extension_penalty = gap_extension_penalty
@@ -90,6 +99,8 @@ class LocalAlignment(object):
         self.verbose = verbose
         self.prefer_gap_runs = prefer_gap_runs
         self.globalalign = globalalign
+        self.wildcard = wildcard
+        self.full_query = full_query
 
     def align(self, ref, query, ref_name='', query_name='', rc=False):
         orig_ref = ref
@@ -99,10 +110,10 @@ class LocalAlignment(object):
         query = query.upper()
 
         matrix = Matrix(len(query) + 1, len(ref) + 1, (0, ' ', 0))
-        for row in xrange(1, matrix.rows):
+        for row in range(1, matrix.rows):
             matrix.set(row, 0, (0, 'i', 0))
 
-        for col in xrange(1, matrix.cols):
+        for col in range(1, matrix.cols):
             matrix.set(0, col, (0, 'd', 0))
 
         max_val = 0
@@ -110,9 +121,9 @@ class LocalAlignment(object):
         max_col = 0
 
         # calculate matrix
-        for row in xrange(1, matrix.rows):
-            for col in xrange(1, matrix.cols):
-                mm_val = matrix.get(row - 1, col - 1)[0] + self.scoring_matrix.score(query[row - 1], ref[col - 1])
+        for row in range(1, matrix.rows):
+            for col in range(1, matrix.cols):
+                mm_val = matrix.get(row - 1, col - 1)[0] + self.scoring_matrix.score(query[row - 1], ref[col - 1], self.wildcard)
 
                 ins_run = 0
                 del_run = 0
@@ -126,10 +137,7 @@ class LocalAlignment(object):
                         if not self.gap_extension_decay:
                             ins_val = matrix.get(row - 1, col)[0] + self.gap_extension_penalty
                         else:
-                            if self.globalalign:
-                                ins_val = matrix.get(row - 1, col)[0] + min(self.gap_extension_penalty + ins_run * self.gap_extension_decay)
-                            else:
-                                ins_val = matrix.get(row - 1, col)[0] + min(0, self.gap_extension_penalty + ins_run * self.gap_extension_decay)
+                            ins_val = matrix.get(row - 1, col)[0] + min(0, self.gap_extension_penalty + ins_run * self.gap_extension_decay)
                 else:
                     ins_val = matrix.get(row - 1, col)[0] + self.gap_penalty
 
@@ -142,15 +150,12 @@ class LocalAlignment(object):
                         if not self.gap_extension_decay:
                             del_val = matrix.get(row, col - 1)[0] + self.gap_extension_penalty
                         else:
-                            if self.globalalign:
-                                del_val = matrix.get(row, col - 1)[0] + min(self.gap_extension_penalty + del_run * self.gap_extension_decay)
-                            else:
-                                del_val = matrix.get(row, col - 1)[0] + min(0, self.gap_extension_penalty + del_run * self.gap_extension_decay)
+                            del_val = matrix.get(row, col - 1)[0] + min(0, self.gap_extension_penalty + del_run * self.gap_extension_decay)
 
                 else:
                     del_val = matrix.get(row, col - 1)[0] + self.gap_penalty
 
-                if self.globalalign:
+                if self.globalalign or self.full_query:
                     cell_val = max(mm_val, del_val, ins_val)
                 else:
                     cell_val = max(mm_val, del_val, ins_val, 0)
@@ -181,10 +186,23 @@ class LocalAlignment(object):
 
         # backtrack
         if self.globalalign:
+            # backtrack from last cell
             row = matrix.rows - 1
             col = matrix.cols - 1
             val = matrix.get(row, col)[0]
+        elif self.full_query:
+            # backtrack from max in last row
+            row = matrix.rows - 1
+            max_val = 0
+            col = 0
+            for c in range(1, matrix.cols):
+                if matrix.get(row, c)[0] > max_val:
+                    col = c
+                    max_val = matrix.get(row, c)[0]
+            col = matrix.cols - 1
+            val = matrix.get(row, col)[0]
         else:
+            # backtrack from max
             row = max_row
             col = max_col
             val = max_val
@@ -199,6 +217,9 @@ class LocalAlignment(object):
             if self.globalalign:
                 if row == 0 and col == 0:
                     break
+            elif self.full_query:
+                if row == 0:
+                    break
             else:
                 if val <= 0:
                     break
@@ -220,23 +241,23 @@ class LocalAlignment(object):
 
         if self.verbose:
             self.dump_matrix(ref, query, matrix, path)
-            print aln
-            print (max_row, max_col), max_val
+            print(aln)
+            print((max_row, max_col), max_val)
 
         cigar = _reduce_cigar(aln)
-        return Alignment(orig_query, orig_ref, row, col, cigar, max_val, ref_name, query_name, rc, self.globalalign)
+        return Alignment(orig_query, orig_ref, row, col, cigar, max_val, ref_name, query_name, rc, self.globalalign, self.wildcard)
 
     def dump_matrix(self, ref, query, matrix, path, show_row=-1, show_col=-1):
         sys.stdout.write('      -      ')
         sys.stdout.write('       '.join(ref))
         sys.stdout.write('\n')
-        for row in xrange(matrix.rows):
+        for row in range(matrix.rows):
             if row == 0:
                 sys.stdout.write('-')
             else:
                 sys.stdout.write(query[row - 1])
 
-            for col in xrange(matrix.cols):
+            for col in range(matrix.cols):
                 if show_row == row and show_col == col:
                     sys.stdout.write('       *')
                 else:
@@ -269,7 +290,7 @@ def _cigar_str(cigar):
 
 
 class Alignment(object):
-    def __init__(self, query, ref, q_pos, r_pos, cigar, score, ref_name='', query_name='', rc=False, globalalign=False):
+    def __init__(self, query, ref, q_pos, r_pos, cigar, score, ref_name='', query_name='', rc=False, globalalign=False, wildcard=None):
         self.query = query
         self.ref = ref
         self.q_pos = q_pos
@@ -280,6 +301,7 @@ class Alignment(object):
         self.q_name = query_name
         self.rc = rc
         self.globalalign = globalalign
+        self.wildcard = wildcard
 
         self.r_offset = 0
         self.r_region = None
@@ -303,7 +325,7 @@ class Alignment(object):
             if op == 'M':
                 q_len += count
                 r_len += count
-                for k in xrange(count):
+                for k in range(count):
                     if self.query[j] == self.ref[i]:
                         self.matches += 1
                     else:
@@ -340,7 +362,7 @@ class Alignment(object):
         working = []
         for count, op in self.cigar:
             if op == 'M':
-                for k in xrange(count):
+                for k in range(count):
                     if self.query[self.q_pos + qpos + k] == self.ref[self.r_pos + rpos + k]:
                         ext_cigar_str += 'M'
                     else:
@@ -380,10 +402,10 @@ class Alignment(object):
             if op == 'M':
                 qlen += count
                 rlen += count
-                for k in xrange(count):
+                for k in range(count):
                     q += self.orig_query[j]
                     r += self.orig_ref[i]
-                    if self.query[j] == self.ref[i]:
+                    if self.query[j] == self.ref[i] or (self.wildcard and (self.query[j] in self.wildcard or self.ref[i] in self.wildcard)):
                         m += '|'
                     else:
                         m += '.'
@@ -392,14 +414,14 @@ class Alignment(object):
                     j += 1
             elif op == 'D':
                 rlen += count
-                for k in xrange(count):
+                for k in range(count):
                     q += '-'
                     r += self.orig_ref[i]
                     m += ' '
                     i += 1
             elif op == 'I':
                 qlen += count
-                for k in xrange(count):
+                for k in range(count):
                     q += self.orig_query[j]
                     r += '-'
                     m += ' '
